--------------------------------------------------------------------------------
-- Copyright (c) 2003 by Trenz Electronic.
-- Duenner Kirchweg 77, 32257 Buende, Germany, www.trenz-electronic.de
--------------------------------------------------------------------------------
-- Project:     TE-XC2S Application Note: Game of Life
-- File:        Life.vhd
-- Description: Control Unit.
-- History:     2001sep01, FB, Created.
--              2003jan09, FB, gameT is std_logic_vector.
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

entity Life is
	port(
		clk:            in  STD_LOGIC;                     -- pixel clock
		rst:            in  STD_LOGIC;                     -- reset
		run:            in  STD_LOGIC;                     -- run/ stop
		gen:            out STD_LOGIC_VECTOR(15 downto 0); -- generation count
		-- VGA interface
		hblk, vblk:     in  STD_LOGIC;                     -- blanking
		row:            in  STD_LOGIC_VECTOR(9 downto 0);  -- scan line
		d:              out STD_LOGIC;                     -- RGB out
		-- RAM interface
		adx:            out STD_LOGIC_VECTOR(11 downto 0); -- address
		rdd_a,  rdd_b:  in  STD_LOGIC_VECTOR(0 downto 0);  -- read data
		wrd:            out STD_LOGIC_VECTOR(0 downto 0);  -- write data
		wren_a, wren_b: out STD_LOGIC                      -- write enable
		);
end Life;

architecture bhv of Life is
	signal   row_g, col_g: UNSIGNED(5 downto 0); -- game (row/col)
	signal   row_d, col_d: UNSIGNED(5 downto 0); -- display (row/col)
	
	signal   bank:         STD_LOGIC;            -- memory bank
	signal   wren:         STD_LOGIC;            -- write enable
	signal   rdd:          STD_LOGIC;            -- read data
	signal   blk:          STD_LOGIC;            -- blanking
	
--	type     gameT is     (start, nw, n, ne, w, e, sw, s, se, write, idle, video);
	subtype  gameT is      std_logic_vector(3 downto 0);
	constant start:        gameT:= "0000";
	constant nw:           gameT:= "0001";
	constant n:            gameT:= "0010";
	constant ne:           gameT:= "0011";
	constant w:            gameT:= "0100";
	constant e:            gameT:= "0101";
	constant sw:           gameT:= "0110";
	constant s:            gameT:= "0111";
	constant se:           gameT:= "1000";
	constant write:        gameT:= "1001";
	constant idle:         gameT:= "1010";
	constant video:        gameT:= "1011";
	signal   game:         gameT;
	
begin           
	------------------------------------
	-- RAM interface
	ram_blk: block
		signal adx_g: STD_LOGIC_VECTOR(adx'range);
		signal adx_d: STD_LOGIC_VECTOR(adx'range);
		begin
		
		-- mux game and display addresses			
		adx_g<= STD_LOGIC_VECTOR(row_g) & STD_LOGIC_VECTOR(col_g);
		adx_d<= STD_LOGIC_VECTOR(row_d) & STD_LOGIC_VECTOR(col_d);
		adx  <= adx_d when blk= '0' or game= idle else adx_g;
		
		-- demux ram data, create write enable
		rdd   <= rdd_a(0) when bank= '0' else rdd_b(0);
		wren_a<= '0'      when bank= '0' else wren;
		wren_b<= wren     when bank= '0' else '0';
	end block;
	
	------------------------------------
	-- VGA interface
	vga_blk: block
		signal col_div: UNSIGNED(3 downto 0);
		begin
		
		blk<= hblk or vblk;
		
		-- create 10x8 superpixel to map into 64x64 memory
		process(clk, rst)
		begin
			if rst= '1' then 
				col_div<= TO_UNSIGNED(0, col_div'length);
				col_d  <= TO_UNSIGNED(0, col_d'length);
			elsif rising_edge(clk) then
				if hblk= '1' then
					col_div<= TO_UNSIGNED(0, col_div'length);
					col_d  <= TO_UNSIGNED(0, col_d'length);
				else
					col_div<= col_div + 1;
					if col_div= TO_UNSIGNED(9, col_div'length) then
						col_div<= TO_UNSIGNED(0, col_div'length);
						col_d<= col_d + 1;
					end if;
				end if;
			end if;
		end process;
		
		row_d<= UNSIGNED(row(8 downto 3));
		
		d<= rdd;
	end block;
	
	------------------------------------
	-- game controller
	game_blk: block
		signal   rowcol:       UNSIGNED(adx'range);
		constant first_rowcol: UNSIGNED(rowcol'range):= (others=> '0');
		signal   row,  col:    UNSIGNED(row_g'range);
		signal   neighbours:   UNSIGNED(3 downto 0);
		signal   game2:        gameT;
		signal   wrd_ii:       STD_LOGIC;
		signal   gen_ii:       UNSIGNED(gen'range);
		
		begin
		row<= rowcol(5 downto 0);
		col<= rowcol(11 downto 6);
		
		-- synchronous RAM usage:
		--     adx is created from next state
		--     drd is assumed to go valid one cycle later
		--
		--      |___     ___     ___           ___     ___     ___     
		-- clk  /   \___/   \___/   \___/ ... /   \___/   \___/   \___/ ...
		--      |_______ _______ _______       _______ _______ _______
		-- game /_start_/__nw___/___n___/ ... /__se___/_write_/_stop__/ ...
		--      |_______ _______       _______ ________ ______ 
		-- adx  /__nw___/___n___/ ... /__se___/__mid__/__mid__/-------- ...
		--      |        _______ _______       _______ _______        
		-- rdd  |-------/__nw___/___n___/ ... /__se___/__mid__/-------- ...
		--      |                                      _______
		-- wrd  |-------------------------------------/__new__/-------- ...
		--      |
		with game select game2<= -- next state
		nw               when start,
		n                when nw,
		ne               when n,
		w                when ne,
		e                when w,
		sw               when e,
		s                when sw,
		se               when s,
		write            when se,
		idle             when write,
		video            when idle,
		start            when video,
		idle             when others; -- safety trap
		
		-- vertical deviation
		with game2 select row_g<=
		row-1          when nw | n            | ne,
		row+0          when w  | write | idle | e,
		row+1          when sw | s            | se,
		(others=> '-') when others;
		
		-- horizontal deviation
		with game2 select col_g<=
		col-1          when nw | w            | sw,
		col+0          when n  | write | idle | s,
		col+1          when ne | e            | se,
		(others=> '-') when others;
		
		-- write enable
		wren  <= '1' when game= write else '0';
		
		-- birth & death rules
		wrd(0)<= wrd_ii; -- when game= write else '-';
		wrd_ii<= rdd when neighbours= 2
		else     '1' when neighbours= 3
		else     '0';
		
		-- state machine
		process(clk, rst)
		begin
			if rst= '1' then
				rowcol<= first_rowcol;
				bank  <= '0';
				game  <= idle;
				gen_ii<= TO_UNSIGNED(0, gen_ii'length);
			elsif rising_edge(clk) then
				case game is
					
					when start=>
					-- start new generation
					-- or continue with next cell
					if (rowcol= first_rowcol) then
						if (vblk= '1') and (run= '1') then
							neighbours<= TO_UNSIGNED(0, neighbours'length);
							bank      <= not(bank);
							gen_ii    <= gen_ii + 1;
							game      <= game2;
						end if;
					else
						neighbours<= TO_UNSIGNED(0, neighbours'length);
						game      <= game2;
					end if;
					
					when nw | n | ne | w | e | sw | s | se=>
					-- count neighbours in all eight directions
					if rdd= '1' then 
						neighbours<= neighbours + 1; 
					end if;
					game<= game2;
					
					when write=>
					-- write back result
					rowcol<= rowcol + 1;
					game  <= game2;
					
					when idle=>
					-- wait for end of blanking
					if blk= '0' then
						game<= game2;
					end if;
					
					when video=> 
					-- wait for next blanking
					if blk= '1' then
						game<= game2;
					end if;
					
					when others=>
					-- safety trap
						game<= game2;
				end case;
			end if;
		end process;
		
		gen<= STD_LOGIC_VECTOR(gen_ii);
	end block;
	
end bhv;

--------------------------------------------------------------------------------
-- end of file
